/* $Id: visir_dfs_dpr.h,v 1.4 2006-10-27 07:21:26 llundin Exp $
 *
 * This file is part of the VISIR Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2006-10-27 07:21:26 $
 * $Revision: 1.4 $
 * $Name: not supported by cvs2svn $
 */

#ifndef VISIR_DFS_DPR_H
#define VISIR_DFS_DPR_H

/* 

   Do not edit this file!

   It was generated automatically with

   ../CPLtest/perl/bddparse.pl ../calibdb-visir/rul/visir.bdd > visir/visir_dfs_dpr.h

*/

IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_CHO_NOD_JIT",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","CHOPNOD","JITTER"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_CHO_NOD_MOS",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","CHOPNOD","MOSAIC"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_CHO_JIT",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","CHOPPING","JITTER"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_CHO_MOS",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","CHOPPING","MOSAIC"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_NOD_JIT",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","NODDING","JITTER"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_NOD_MOS",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","NODDING","MOSAIC"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_OBS_DIR_JIT",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s %s","IMAGE","DIRECT","JITTER"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_DARK",(catg,"%s","CALIB"),(type,"%s","DARK"),(tech,"%s","IMAGE"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_PHOT",(catg,"%s","CALIB"),(type,"%s","STD"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_FLAT",(catg,"%s","CALIB"),(type,"%s","FLAT"),(tech,"%s %s","IMAGE","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_TECH_FLAT",(catg,"%s","TECHNICAL"),(type,"%s","FLAT"),(tech,"%s %s","IMAGE","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_ILLU",(catg,"%s","CALIB"),(type,"%s","FLUX"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_PSF",(catg,"%s","CALIB"),(type,"%s","PSF"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_PFOV_BIN",(catg,"%s","TECHNICAL"),(type,"%s %s","OBJECT","PFOV"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_PFOV_TEL",(catg,"%s","TECHNICAL"),(type,"%s %s","STD","PFOV"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_TEC_FOCUS",(catg,"%s","TECHNICAL"),(type,"%s","FOCUS"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_TEC_ALIGN",(catg,"%s","TECHNICAL"),(type,"%s %s","STD","ALIGNMENT"),(tech,"%s %s","IMAGE","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_TEC_TRANS",(catg,"%s","TECHNICAL"),(type,"%s %s","STD","TRANSMISSION"),(tech,"%s %s","IMAGE","CHOPPING"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_ACHRO",(catg,"%s","TECHNICAL"),(type,"%s","OBJECT"),(tech,"%s %s","IMAGE","CHOPPING"));
IRPLIB_DFS_DOCATG_FROM_DPR("IM_CAL_FLEX",(catg,"%s","TECHNICAL"),(type,"%s %s","STD","FLEXURE"),(tech,"%s %s","IMAGE","CHOPPING"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_OBS_LMR",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s","SPECTRUM","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_OBS_HR",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s %s","SPECTRUM","NODDING"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_OBS_HRG",(catg,"%s","SCIENCE"),(type,"%s","OBJECT"),(tech,"%s","ECHELLE"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_DARK",(catg,"%s","CALIB"),(type,"%s","DARK"),(tech,"%s","SPECTRUM"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_LMR_WCAL",(catg,"%s","CALIB"),(type,"%s","WAVE"),(tech,"%s %s","SPECTRUM","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_HRG_WCAL",(catg,"%s","CALIB"),(type,"%s","WAVE"),(tech,"%s","ECHELLE"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_PHOT_HRG",(catg,"%s","CALIB"),(type,"%s","STD"),(tech,"%s","ECHELLE"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_PHOT",(catg,"%s","CALIB"),(type,"%s","STD"),(tech,"%s %s","SPECTRUM","CHOPNOD"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_FLAT",(catg,"%s","CALIB"),(type,"%s","FLAT"),(tech,"%s %s","SPECTRUM","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_TECH_FLAT",(catg,"%s","TECHNICAL"),(type,"%s","FLAT"),(tech,"%s %s","SPECTRUM","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_SLIT",(catg,"%s","TECHNICAL"),(type,"%s","SLIT"),(tech,"%s %s","SPECTRUM","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_DIST_SKY",(catg,"%s","CALIB"),(type,"%s","SKY"),(tech,"%s %s","SPECTRUM","DIRECT"));
IRPLIB_DFS_DOCATG_FROM_DPR("SPEC_CAL_DIST_SPEC",(catg,"%s","CALIB"),(type,"%s","LAMP"),(tech,"%s %s","SPECTRUM","CHOPPING"));

#endif
